use crate::mailuser::Mailuser;
use std::collections::{BTreeMap, HashMap};

#[cfg(feature = "serde")]
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone)]
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
pub struct Certificate {
    pub der: Vec<u8>,
    pub key: Vec<u8>,
}

/// An item which stores known certificates. For convenience, this trait is implemented for
/// `std::collections::HashMap<String, String>` and `std::collections::BTreeMap<String, String>`
pub trait CertificateStore: Send + Sync {
    /// Retrieves a certificate fingerprint using the hostname as it's key if
    /// it has previously been saved
    fn get_certificate(&self, host: &str) -> Option<String>;
    /// Inserts a certificate fingerprint with the hostname as the key
    fn insert_certificate(&mut self, host: &str, fingerprint: &str) -> Option<String>;
    /// Returns true if a fingerprint has been stored for this host
    fn contains_certificate(&self, host: &str) -> bool;
}

/// An item which stores client certificates. For convenience, this trait is implemented for
/// `std::collections::HashMap<String, Certificate>` and `std::collections::BTreeMap<String, Certificate>`
pub trait ClientCertificateStore: Send + Sync {
    /// Retrieves a client certificate using the sender's address as the key
    fn get_certificate(&self, user: &Mailuser) -> Option<Certificate>;
    /// Inserts a client certificate with the sender's address as the key
    fn insert_certificate(
        &mut self,
        user: &Mailuser,
        certificate: Certificate,
    ) -> Option<Certificate>;
    /// Returns true if the sender has a certificate in this store
    fn contains_certificate(&self, user: &Mailuser) -> bool;
}

impl CertificateStore for HashMap<String, String> {
    fn get_certificate(&self, host: &str) -> Option<String> {
        self.get(host).cloned()
    }

    fn insert_certificate(&mut self, host: &str, fingerprint: &str) -> Option<String> {
        self.insert(host.to_string(), fingerprint.to_string())
    }

    fn contains_certificate(&self, host: &str) -> bool {
        self.contains_key(host)
    }
}

impl CertificateStore for BTreeMap<String, String> {
    fn get_certificate(&self, host: &str) -> Option<String> {
        self.get(host).cloned()
    }

    fn insert_certificate(&mut self, host: &str, fingerprint: &str) -> Option<String> {
        self.insert(host.to_string(), fingerprint.to_string())
    }

    fn contains_certificate(&self, host: &str) -> bool {
        self.contains_key(host)
    }
}

impl ClientCertificateStore for HashMap<String, Certificate> {
    fn get_certificate(&self, user: &Mailuser) -> Option<Certificate> {
        self.get(&user.to_string()).cloned()
    }

    fn insert_certificate(
        &mut self,
        user: &Mailuser,
        certificate: Certificate,
    ) -> Option<Certificate> {
        self.insert(user.to_string(), certificate)
    }

    fn contains_certificate(&self, user: &Mailuser) -> bool {
        self.contains_key(&user.to_string())
    }
}

impl ClientCertificateStore for BTreeMap<String, Certificate> {
    fn get_certificate(&self, user: &Mailuser) -> Option<Certificate> {
        self.get(&user.to_string()).cloned()
    }

    fn insert_certificate(
        &mut self,
        user: &Mailuser,
        certificate: Certificate,
    ) -> Option<Certificate> {
        self.insert(user.to_string(), certificate)
    }

    fn contains_certificate(&self, user: &Mailuser) -> bool {
        self.contains_key(&user.to_string())
    }
}
