use std::{
    fmt,
    net::{SocketAddr, ToSocketAddrs},
    str::FromStr,
};
mod error;
pub use error::Error;

#[cfg(feature = "serde")]
use serde::{Deserialize, Serialize};

#[derive(Clone, Debug, Default, PartialEq)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
/// Represents the fully qualified domain name for this host
/// # Examples
/// Parse a host from a given string
/// ```
/// use dory::prelude::Host;
///
/// let host_str = "misfin.example.com";
/// let host: Host = host_str.parse().unwrap();
/// assert_eq!(host.subdomain.unwrap().as_str(), "misfin");
/// assert_eq!(host.domain.as_str(), "example");
/// assert_eq!(host.tld.as_str(), "com");
/// ```
pub struct Host {
    pub subdomain: Option<String>,
    pub domain: String,
    pub tld: String,
}

impl fmt::Display for Host {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(ref sub) = self.subdomain {
            write!(f, "{sub}.{}.{}", self.domain, self.tld)
        } else {
            write!(f, "{}.{}", self.domain, self.tld)
        }
    }
}

impl FromStr for Host {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if s.contains(char::is_whitespace) {
            return Err(Error::IllegalWhitespace);
        }
        if let Some((domain, tld)) = s.rsplit_once('.') {
            if domain.is_empty() {
                Err(Error::EmptyDomain)
            } else if tld.is_empty() {
                Err(Error::EmptyTld)
            } else if let Some((subdomain, domain)) = domain.rsplit_once('.') {
                if subdomain.is_empty() {
                    Err(Error::EmptySubdomain)
                } else if domain.is_empty() {
                    Err(Error::EmptyDomain)
                } else {
                    Ok(Host {
                        subdomain: Some(subdomain.to_string()),
                        domain: domain.to_string(),
                        tld: tld.to_string(),
                    })
                }
            } else {
                Ok(Host {
                    subdomain: None,
                    domain: domain.to_string(),
                    tld: tld.to_string(),
                })
            }
        } else {
            Err(Error::MissingSeparator)
        }
    }
}

impl ToSocketAddrs for Host {
    type Iter = std::vec::IntoIter<SocketAddr>;

    fn to_socket_addrs(&self) -> std::io::Result<Self::Iter> {
        format!("{self}:1958").to_socket_addrs()
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn parse_missing_separator() {
        assert_eq!(
            "exampledotcom".parse::<Host>(),
            Err(Error::MissingSeparator)
        );
    }

    #[test]
    fn parse_empty_tld() {
        assert_eq!("example.".parse::<Host>(), Err(Error::EmptyTld));
    }

    #[test]
    fn parse_empty_domain() {
        assert_eq!(".com".parse::<Host>(), Err(Error::EmptyDomain));
        assert_eq!("example..com".parse::<Host>(), Err(Error::EmptyDomain));
    }

    #[test]
    fn parse_empty_subdomain() {
        assert_eq!(".example.com".parse::<Host>(), Err(Error::EmptySubdomain));
    }

    #[test]
    fn parse_illegal_whitespace() {
        assert_eq!(
            "exam\tple.com".parse::<Host>(),
            Err(Error::IllegalWhitespace)
        );
    }

    #[test]
    fn parse_host() {
        assert!("example.com".parse::<Host>().is_ok());
    }

    #[test]
    fn parse_host_with_subdomain() {
        let host: Host = "mail.example.com".parse().unwrap();
        assert_eq!(host.subdomain, Some(String::from("mail")));
    }

    #[test]
    fn parse_complex_subdomain() {
        let host: Host = "mail.sender.example.com".parse().unwrap();
        assert_eq!(host.subdomain, Some(String::from("mail.sender")));
    }

    #[test]
    fn fmt() {
        let host = Host {
            subdomain: None,
            domain: String::from("example"),
            tld: String::from("com"),
        };
        let host_str = host.to_string();
        assert_eq!(host_str.as_str(), "example.com");
    }

    #[test]
    fn fmt_with_subdomain() {
        let host = Host {
            subdomain: Some(String::from("mail")),
            domain: String::from("example"),
            tld: String::from("com"),
        };
        let host_str = host.to_string();
        assert_eq!(host_str.as_str(), "mail.example.com");
    }
}
