use crate::prelude::{Mailbox, Mailuser};
use std::{fmt, str::FromStr};

#[cfg(feature = "serde")]
use serde::{Deserialize, Serialize};

mod error;
pub use error::Error;

#[derive(Clone, Debug, PartialEq)]
#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
/// The full request as sent by the `Sender` and received by the `Receiver`
pub struct Request {
    /// The sender of the message
    pub recipient: Mailuser,
    /// The message body
    pub message: String,
}

impl fmt::Display for Request {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "misfin://{} {}\r\n", self.recipient, self.message)
    }
}

impl FromStr for Request {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Some((user, message)) = s.split_once(' ') {
            let Some(message) = message.strip_suffix("\r\n").map(ToString::to_string) else {
                return Err(Error::Malformed);
            };
            if message.is_empty() {
                return Err(Error::EmptyMessage);
            }
            let Some(user) =  user.strip_prefix("misfin://") else {
                return Err(Error::Malformed);
            };
            let recipient = user.parse()?;
            Ok(Request { recipient, message })
        } else {
            Err(Error::MissingSeparator)
        }
    }
}

impl Request {
    pub fn recipients(&self) -> Vec<Mailuser> {
        let mut recipients = vec![];
        self.message.lines().for_each(|l| {
            if let Some(s) = l.strip_prefix(':') {
                s.split_whitespace().for_each(|u| {
                    if let Ok(user) = u.parse() {
                        recipients.push(user);
                    }
                });
            }
        });
        recipients
    }

    pub fn senders(&self) -> Vec<Mailbox> {
        let mut senders = vec![];
        self.message.lines().for_each(|l| {
            if let Some(s) = l.strip_prefix('<') {
                if let Ok(mbox) = s.trim().parse() {
                    senders.push(mbox);
                }
            }
        });
        senders
    }

    pub fn timestamp(&self) -> Option<String> {
        for l in self.message.lines() {
            if let Some(s) = l.strip_prefix('@') {
                return Some(s.trim().to_string());
            }
        }
        None
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::prelude::{Host, ParseHostError};

    const REQ_STR: &'static str = "misfin://john@misfin.example.com Anyone seen Jane?\r\n";
    const REQ_COMPLEX: &'static str = "misfin://john@misfin.example.com \
        < jane@misfin.example.xyz Jane Doe\n\
        <bob@gemini.pizza Bob Thornton\n\
        : bruce@willis.abc demi@moore.com billy@ray.cyrus\n\
        @ 2023-05-09T19:39:15Z\n\
        Hello world!\r\n";

    fn req() -> Request {
        Request {
            recipient: Mailuser {
                username: "john".to_string(),
                host: Host {
                    subdomain: Some("misfin".into()),
                    domain: "example".into(),
                    tld: "com".into(),
                },
            },
            message: "Anyone seen Jane?".to_string(),
        }
    }

    #[test]
    fn to_string() {
        assert_eq!(req().to_string().as_str(), REQ_STR,);
    }

    #[test]
    fn parse() {
        assert_eq!(REQ_STR.parse::<Request>().unwrap(), req());
    }

    #[test]
    fn parse_missing_sep() {
        let req = "misfin://john@example.comHelloWorld!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::MissingSeparator));
    }

    #[test]
    fn parse_malformed() {
        let req = "misfin://john@example.com Hello World!\n".parse::<Request>();
        assert_eq!(req, Err(Error::Malformed));
        let req = "mail://john@example.com Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::Malformed));
    }

    #[test]
    fn parse_empty_user() {
        let req = "misfin://@example.com Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::EmptyUser));
    }

    #[test]
    fn parse_empty_host() {
        let req = "misfin://john@ Hello World!\r\n".parse::<Request>();
        assert_eq!(req, Err(Error::EmptyHost));
    }

    #[test]
    fn parse_empty_msg() {
        let req = "misfin://john@example.com \r\n".parse::<Request>();
        assert_eq!(req, Err(Error::EmptyMessage));
    }

    #[test]
    fn parse_host_illegal_whitespace() {
        let req = "misfin://john@example\tfairy.com Hello World!\r\n".parse::<Request>();
        assert_eq!(
            req,
            Err(Error::ParseHostError(ParseHostError::IllegalWhitespace))
        );
    }

    #[test]
    fn recipients() {
        let req = REQ_COMPLEX.parse::<Request>().unwrap();
        let recipients = req.recipients();
        assert_eq!(recipients.len(), 3);
        let first = recipients.iter().next().unwrap();
        assert_eq!(first.username, "bruce");
        assert!(first.host.subdomain.is_none());
        assert_eq!(first.host.domain, "willis");
        assert_eq!(first.host.tld, "abc");
    }

    #[test]
    fn senders() {
        let req = REQ_COMPLEX.parse::<Request>().unwrap();
        let senders = req.senders();
        assert_eq!(senders.len(), 2);
        let last = senders.iter().last().unwrap();
        assert_eq!(last.host.tld, "pizza");
    }

    #[test]
    fn timestamp() {
        let req = REQ_COMPLEX.parse::<Request>().unwrap();
        let ts = req.timestamp().unwrap();
        assert_eq!(ts, "2023-05-09T19:39:15Z")
    }
}
