use {
    crate::prelude::ParseStatusError,
    std::{fmt, num::ParseIntError, string::FromUtf8Error},
};

#[derive(Debug, PartialEq)]
/// Errors which can occur when parsing the response sent by the receving server
/// back to the sender
pub enum Error {
    /// The message was too long. The Misfin spec allows a maximum length of
    /// 2048 bytes for the message body.
    TooLong,
    /// An error occurred parsing the status code as a number.
    ParseInt(ParseIntError),
    /// The server sent an invalid or unrecognized status code
    StatusError,
    /// The response was malformed
    Malformed,
    /// The response is not valid utf8
    Utf8Error,
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::TooLong => write!(f, "ParseResponseError: too long"),
            Self::ParseInt(e) => write!(f, "ParseResponseError: {e}"),
            Self::StatusError => write!(f, "ParseResponseError: Invalid Status"),
            Self::Malformed => write!(f, "ParseResponseError: Malformed"),
            Self::Utf8Error => write!(f, "ParseResponseError: Not Utf8"),
        }
    }
}

impl std::error::Error for Error {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::ParseInt(e) => Some(e),
            _ => None,
        }
    }
}

impl From<ParseIntError> for Error {
    fn from(value: ParseIntError) -> Self {
        Self::ParseInt(value)
    }
}

impl From<ParseStatusError> for Error {
    fn from(_value: ParseStatusError) -> Self {
        Self::StatusError
    }
}

impl From<FromUtf8Error> for Error {
    fn from(_value: FromUtf8Error) -> Self {
        Self::Utf8Error
    }
}
