use {
    crate::prelude::{ParseMailboxError, ParseRequestError, ParseResponseError},
    rustls::InvalidMessage,
    std::{fmt, io},
};

#[derive(Debug)]
/// Errors which might occur when sending a message
pub enum Error {
    CertificateError(InvalidMessage),
    DnsError,
    IoError(io::Error),
    MailboxError(ParseMailboxError),
    RequestError(ParseRequestError),
    ResponseError(ParseResponseError),
    TlsError(rustls::Error),
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::CertificateError(e) => write!(f, "{e:?}"),
            Self::DnsError => write!(f, "Dns Error"),
            Self::IoError(e) => write!(f, "{e}"),
            Self::MailboxError(e) => write!(f, "{e}"),
            Self::RequestError(e) => write!(f, "{e}"),
            Self::ResponseError(e) => write!(f, "{e}"),
            Self::TlsError(e) => write!(f, "{e}"),
        }
    }
}

impl std::error::Error for Error {
    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
        match self {
            Self::IoError(e) => Some(e),
            Self::MailboxError(e) => Some(e),
            Self::RequestError(e) => Some(e),
            Self::ResponseError(e) => Some(e),
            Self::TlsError(e) => Some(e),
            _ => None,
        }
    }
}

impl From<InvalidMessage> for Error {
    fn from(value: InvalidMessage) -> Self {
        Self::CertificateError(value)
    }
}

impl From<rustls::Error> for Error {
    fn from(value: rustls::Error) -> Self {
        Self::TlsError(value)
    }
}

impl From<ParseMailboxError> for Error {
    fn from(value: ParseMailboxError) -> Self {
        Self::MailboxError(value)
    }
}

impl From<ParseRequestError> for Error {
    fn from(value: ParseRequestError) -> Self {
        Self::RequestError(value)
    }
}

impl From<ParseResponseError> for Error {
    fn from(value: ParseResponseError) -> Self {
        Self::ResponseError(value)
    }
}

impl From<io::Error> for Error {
    fn from(value: io::Error) -> Self {
        Self::IoError(value)
    }
}
