use {
    crate::prelude::{CertificateStore, GetFingerprint},
    rustls::{
        client::{ServerCertVerified, ServerCertVerifier},
        Certificate,
    },
    std::{borrow::BorrowMut, sync::Mutex, time},
};

#[derive(Debug)]
/// A verifier is used to verify certificates sent by the receiving server
/// during the tls handshake. For convenience the [`CertificateStore`] trait
/// is implemented for [`std::collections::HashMap<String, String>`] and
/// [`std::collections::BTreeMap<String, String>`], so a Verifier can easily
/// be constructed from those types.
///
/// This type is passed to [`rustls::client::DangerousClientConfig::set_certificate_verifier`]
/// in order to allow Tofu certificate validation.
/// # Examples
/// Create a Verifier from a `HashMap`
/// ```
/// use std::collections::HashMap;
/// use dory::prelude::{CertificateStore, Verifier};
///
/// let verifier: Verifier<HashMap<String, String>> = HashMap::new().into();
/// ```
pub struct Verifier<S: CertificateStore> {
    /// An item which serves as storage for certificates
    pub store: Mutex<S>,
}

impl<S: CertificateStore> ServerCertVerifier for Verifier<S> {
    fn verify_server_cert(
        &self,
        end_entity: &Certificate,
        _intermediates: &[Certificate],
        server_name: &rustls::ServerName,
        _scts: &mut dyn Iterator<Item = &[u8]>,
        _ocsp_response: &[u8],
        _now: time::SystemTime,
    ) -> Result<ServerCertVerified, rustls::Error> {
        let fp = end_entity
            .fingerprint()
            .map_err(|e| rustls::Error::General(e.to_string()))?;
        let name = match server_name {
            rustls::ServerName::DnsName(n) => n.as_ref().to_string(),
            rustls::ServerName::IpAddress(ip) => ip.to_string(),
            _ => todo!(),
        };
        let mut store = self.store.lock().unwrap();
        if let Some(fingerprint) = store.get_certificate(&name) {
            // TODO: needs a lot more checking for certificate validity
            if fingerprint == fp.fingerprint {
                for n in fp.names {
                    if n == name {
                        return Ok(ServerCertVerified::assertion());
                    }
                }
                Err(rustls::Error::InvalidCertificate(
                    rustls::CertificateError::NotValidForName,
                ))
            } else {
                Err(rustls::Error::InvalidCertificate(
                    rustls::CertificateError::NotValidForName,
                ))
            }
        } else {
            if !store.contains_certificate(&name) {
                let _key = store
                    .borrow_mut()
                    .insert_certificate(&name, &fp.fingerprint);
            }
            Ok(ServerCertVerified::assertion())
        }
    }
}

impl<T: CertificateStore> From<T> for Verifier<T> {
    fn from(value: T) -> Self {
        Self {
            store: Mutex::new(value),
        }
    }
}

impl<T: CertificateStore> Verifier<T> {
    pub fn new(store: T) -> Self {
        store.into()
    }
}

#[cfg(test)]
mod tests {
    use std::collections::HashMap;

    use super::*;

    const CERT: &[u8] = include_bytes!("../../test/certificates/gemini.example.org/cert.der");
    const FP: &'static str =
        include_str!("../../test/certificates/gemini.example.org/fingerprint.txt");

    #[test]
    fn tofu() {
        let verifier: Verifier<HashMap<String, String>> = HashMap::new().into();
        let cert = rustls::Certificate(CERT.into());
        let name = rustls::ServerName::try_from("gemini.example.org").unwrap();
        let now = time::SystemTime::now();
        let scts: Vec<&[u8]> = vec![];
        let res =
            verifier.verify_server_cert(&cert, &[], &name, &mut scts.iter().copied(), &[], now);
        assert!(res.is_ok());
    }

    #[test]
    fn verify_previously_trusted() {
        let verifier: Verifier<HashMap<String, String>> =
            HashMap::from([("gemini.example.org".to_string(), FP.to_string())]).into();
        let cert = rustls::Certificate(CERT.into());
        let name = rustls::ServerName::try_from("gemini.example.org").unwrap();
        let now = time::SystemTime::now();
        let scts: Vec<&[u8]> = vec![];
        let res =
            verifier.verify_server_cert(&cert, &[], &name, &mut scts.iter().copied(), &[], now);
        assert!(res.is_ok());
    }
}
